module ioput_module
use amr_module
use amrray_module
use stars_module
use dust_module
use rtglobal_module
use constants_module

!character*160 :: gridfilename
!integer :: gridfilenamelen

contains



!-------------------------------------------------------------------------
!                        READ DUST DENSITY
!-------------------------------------------------------------------------
subroutine read_dust_density(action)
  implicit none
  integer :: action
  logical :: fex1,fex2
  double precision, allocatable :: data(:)
  double precision :: dummy
  integer(kind=8) :: iiformat,reclen,reclend,nn,kk
  integer :: iformat,index,ierr,irec,ispec,k,n,i
  !
  ! Action=0 means do nothing, action=1 means read if not yet read,
  ! action=2 means re-read.
  !
  if(action.eq.0) then
     return
  elseif(action.eq.1) then
     if(allocated(dustdens)) return
  endif
  !
  ! Now read in the density, either from a special file or from the standard
  ! dust_density.inp
  !
  inquire(file='dust_density.inp',exist=fex1)
  inquire(file='dust_density.uinp',exist=fex2)
  if(fex1.and.fex2) then
     write(stdo,*) 'ERROR: Found both dust_density.inp and dust_density.uinp'
     stop
  endif
  if((.not.fex1).and.(.not.fex2)) then
     write(stdo,*) 'ERROR: Could not find the dust_density.inp or dust_density.uinp file...'
     stop
  endif
  write(stdo,*) 'Reading dust densities...'
  call flush(stdo)
  if(fex1) then
     !
     ! Formatted input
     !
     open(unit=1,file='dust_density.inp',status='old')
     read(1,*) iformat
     if(iformat.ne.1) then
        write(stdo,*) 'ERROR: Format number of dust_density.inp is invalid/unknown.'
        stop
     endif
     read(1,*) n
     read(1,*) k
     if(n.ne.nrcellsinp) then
        write(stdo,*) 'ERROR: dust_density.inp does not have same number'
        write(stdo,*) '       of cells as the grid.'
        write(stdo,*) n,nrcellsinp
        stop
     endif
     if(k.ne.dust_nr_species) then
        write(stdo,*) 'ERROR: In the dust_density.inp file the number'
        write(stdo,*) '  of dust species is different from that '
        write(stdo,*) '  specified in dustopac.inp'
        write(stdo,*) k,dust_nr_species
        stop
     endif
     allocate(dustdens(1:dust_nr_species,1:nrcells),STAT=ierr)
     if(ierr.ne.0) then
        write(stdo,*) 'ERROR: Could not allocate dust density array'
        stop
     endif
     allocate(dust_massdust(1:dust_nr_species),STAT=ierr)
     if(ierr.ne.0) then
        write(stdo,*) 'ERROR: Could not allocate dust mass array'
        stop
     endif
     do ispec=1,dust_nr_species
        call amr_resetcount()
        call amr_nextcell(index)
        do while(index.ge.0) 
           read(1,*) dummy 
           if(index.gt.0) dustdens(ispec,index) = max(dummy,1d-99)
           call amr_nextcell(index)
        enddo
     enddo
  else
     !
     ! Unformatted input
     !
     open(unit=1,file='dust_density.uinp',status='old',form='unformatted')
     read(1) iiformat,reclen
     if(iiformat.ne.1) then
        write(stdo,*) 'ERROR: Format number of dust_density.uinp is invalid/unknown.'
        write(stdo,*) 'Format number = ',iiformat
        write(stdo,*) 'Record length = ',reclen
        stop
     endif
     reclend=reclen/8
     read(1) nn,kk
     if(nn.ne.nrcellsinp) then
        write(stdo,*) 'ERROR: dust_density.inp does not have same number'
        write(stdo,*) '       of cells as the grid.'
        write(stdo,*) nn,nrcellsinp
        stop
     endif
     if(kk.ne.dust_nr_species) then
        write(stdo,*) 'ERROR: In the dust_density.uinp file the number'
        write(stdo,*) '  of dust species is different from that '
        write(stdo,*) '  specified in dustopac.inp'
        write(stdo,*) kk,dust_nr_species
        stop
     endif
     allocate(dustdens(1:dust_nr_species,1:nrcells),STAT=ierr)
     if(ierr.ne.0) then
        write(stdo,*) 'ERROR: Could not allocate dust density array'
        stop
     endif
     allocate(dust_massdust(1:dust_nr_species),STAT=ierr)
     if(ierr.ne.0) then
        write(stdo,*) 'ERROR: Could not allocate dust mass array'
        stop
     endif
     if(allocated(data)) deallocate(data)
     allocate(data(1:reclend),STAT=ierr)
     if(ierr.ne.0) then
        write(stdo,*) 'ERROR: Could not allocate data array'
        stop
     endif
     do ispec=1,dust_nr_species
        call amr_resetcount()
        call amr_nextcell(index)
        do irec=1,(nrcellsinp-1)/reclend+1
           read(1) data
           i = 1
           do while((index.ge.0).and.(i.le.reclend))
              if(index.gt.0) dustdens(ispec,index) = max(data(i),1d-99)
              i     = i + 1
              call amr_nextcell(index)
           enddo
        enddo
     enddo
  endif
  close(1)
  if(allocated(data)) deallocate(data)
  !
  ! Compute mass in each dust species and print
  !
  call compute_dust_mass()
  if(dust_nr_species.gt.1) then
     do ispec=1,dust_nr_species
        write(stdo,*) 'Dust mass in species ',ispec,' = ',     &
             dust_massdust(ispec)/1.99d33,' Msun'
     enddo
  endif
  write(stdo,*) 'Dust mass total = ',                      &
       dust_massdusttot/1.99d33,' Msun'
  !
end subroutine read_dust_density



!-------------------------------------------------------------------------
!                       READ DUST TEMPERATURE 
!-------------------------------------------------------------------------
subroutine read_dust_temperature(action)
  implicit none
  integer :: action
  logical :: fex1,fex2
  double precision, allocatable :: data(:)
  double precision :: dummy
  integer(kind=8) :: iiformat,reclen,reclend,nn,kk
  integer :: n,k,i,irec,ispec,index,ierr,iformat
  !
  ! Action=0 means do nothing, action=1 means read if not yet read,
  ! action=2 means re-read.
  !
  if(action.eq.0) then
     return
  elseif(action.eq.1) then
     if(allocated(dusttemp)) return
  endif
  !
  ! First check that the user has not accidently created a 
  ! .inp or .uinp file. This is a common confusion and may lead to
  ! a lot of bug hunting for nothing. The reason why this file is
  ! a .dat and not .inp file is because normally this is a file that
  ! is created by RADMC-3D by the Monte Carlo simulation, and later 
  ! re-read by RADMC-3D for the imaging/spectra. It is therefore a mix
  ! of .inp and .out file, and we thus call it a .dat (or .udat) file.
  !
  inquire(file='dust_temperature.inp',exist=fex1)
  inquire(file='dust_temperature.uinp',exist=fex2)
  if(fex1.or.fex2) then
     write(stdo,*) 'ERROR: Found dust_temperature.inp or dust_temperature.uinp'
     write(stdo,*) '       file. But dust_temperature must be a .dat or .udat file.'
     stop
  endif
  !
  ! Open temperature file
  !
  inquire(file='dust_temperature.dat',exist=fex1)
  inquire(file='dust_temperature.udat',exist=fex2)
  if(fex1.and.fex2) then
     write(stdo,*) 'ERROR: Found both dust_temperature.dat and dust_temperature.udat'
     stop
  endif
  if((.not.fex1).and.(.not.fex2)) then
     write(stdo,*) 'ERROR: Could not find the dust_temperature.dat or dust_temperature.udat file...'
     stop
  endif
  write(stdo,*) 'Reading dust temperatures...'
  call flush(stdo)
  if(fex1) then
     open(unit=1,file='dust_temperature.dat',status='old')
     read(1,*) iformat
     if(iformat.ne.1) then
        write(stdo,*) 'ERROR: Format number of dust_temperature.inp is invalid/unknown.'
        stop
     endif
     !read(1,*) gridfilenamedum
     read(1,*) n
     read(1,*) k
     if(n.ne.nrcellsinp) then
        write(stdo,*) 'ERROR: dust_temperature.inp does not have same number'
        write(stdo,*) '       of cells as the grid.'
        write(stdo,*) n,nrcellsinp
        stop
     endif
     if(k.ne.dust_nr_species) then
        write(stdo,*) 'ERROR: In the dust_temperature.inp file the number'
        write(stdo,*) '  of dust species is different from that '
        write(stdo,*) '  specified in dustopac.inp'
        write(stdo,*) k,dust_nr_species
        stop
     endif
     allocate(dusttemp(1:dust_nr_species,1:nrcells),STAT=ierr)
     if(ierr.ne.0) then
        write(stdo,*) 'ERROR: Could not allocate dust temperature array'
        stop
     endif
     do ispec=1,dust_nr_species
        call amr_resetcount()
        call amr_nextcell(index)
        do while(index.ge.0)
           read(1,*) dummy 
           if(dummy.lt.0) then 
              write(stdo,*) 'ERROR: Negative temperature detected in input file.'
              stop
           endif
           if(dummy.lt.1d-90) then 
              dummy = 1d-90
           endif
           if(index.gt.0) dusttemp(ispec,index) = dummy
           call amr_nextcell(index)
        enddo
     enddo
     close(1)
  else
     !
     ! Unformatted input of dust temperature
     !
     open(unit=1,file='dust_temperature.udat',status='old',form='unformatted')
     read(1) iiformat,reclen
     if(iiformat.ne.1) then
        write(stdo,*) 'ERROR: Format number of dust_temperature.udat is invalid/unknown.'
        stop
     endif
     reclend=reclen/8
     read(1) nn,kk
     if(nn.ne.nrcellsinp) then
        write(stdo,*) 'ERROR: dust_temperature.udat does not have same number'
        write(stdo,*) '       of cells as the grid.'
        write(stdo,*) nn,nrcellsinp
        stop
     endif
     if(kk.ne.dust_nr_species) then
        write(stdo,*) 'ERROR: In the dust_temperature.udat file the number'
        write(stdo,*) '  of dust species is different from that '
        write(stdo,*) '  specified in dustopac.inp'
        write(stdo,*) kk,dust_nr_species
        stop
     endif
     allocate(dusttemp(1:dust_nr_species,1:nrcells),STAT=ierr)
     if(ierr.ne.0) then
        write(stdo,*) 'ERROR: Could not allocate dust temperature array'
        stop
     endif
     if(allocated(data)) deallocate(data)
     allocate(data(1:reclend),STAT=ierr)
     if(ierr.ne.0) then
        write(stdo,*) 'ERROR: Could not allocate data array'
        stop
     endif
     do ispec=1,dust_nr_species
        call amr_resetcount()
        call amr_nextcell(index)
        do irec=1,(nrcellsinp-1)/reclend+1
           read(1) data
           i = 1
           do while((index.ge.0).and.(i.le.reclend))
              if(data(i).lt.1d-90) then 
                 data(i) = 1d-90
              endif
              if(index.gt.0) dusttemp(ispec,index) = data(i) 
              i     = i + 1
              call amr_nextcell(index)
           enddo
        enddo
     enddo
  endif
end subroutine read_dust_temperature


!-------------------------------------------------------------------------
!                       READ GAS DENSITY
!-------------------------------------------------------------------------
subroutine read_gas_density(action)
  implicit none
  integer :: action
  logical :: fex1,fex2
  double precision, allocatable :: data(:)
  double precision :: dummy
  integer(kind=8) :: iiformat,reclen,reclend,nn,kk
  integer :: i,index,iformat,irec,n,k,ierr
  !
  ! Action=0 means do nothing, action=1 means read if not yet read,
  ! action=2 means re-read.
  !
  if(action.eq.0) then
     return
  elseif(action.eq.1) then
     if(allocated(gasdens)) return
  endif
  !
  ! Open temperature file
  !
  inquire(file='gas_density.inp',exist=fex1)
  inquire(file='gas_density.uinp',exist=fex2)
  if(fex1.and.fex2) then
     write(stdo,*) 'ERROR: Found both gas_density.inp and gas_density.uinp'
     stop
  endif
  if(fex1.or.fex2) then
     !
     ! Allocate array
     !
     if(allocated(gasdens)) deallocate(gasdens)
     allocate(gasdens(1:nrcells),STAT=ierr)
     if(ierr.ne.0) then
        write(stdo,*) 'ERROR: Could not allocate gasdens array'
        stop
     endif
     write(stdo,*) 'Reading gas density...'
     call flush(stdo)
     !
     ! Now read this temperature
     !
     if(fex1) then
        !
        ! Formatted gas temperature file
        !
        open(unit=1,file='gas_density.inp',status='old')
        read(1,*) iformat
        if(iformat.ne.1) then
           write(stdo,*) 'ERROR: Format number of gas_density.inp is invalid/unknown.'
           stop
        endif
        !read(1,*) gridfilenamedum
        read(1,*) n
        if(n.ne.nrcellsinp) then
           write(stdo,*) 'ERROR: gas_density.inp does not have same number'
           write(stdo,*) '       of cells as the grid.'
           write(stdo,*) n,nrcellsinp
           stop
        endif
        call amr_resetcount()
        call amr_nextcell(index)
        do while(index.ge.0)
           read(1,*) dummy 
           if(dummy.lt.0) then 
              write(stdo,*) 'ERROR: Negative gas density detected in input file.'
              stop
           endif
           if(dummy.lt.1d-90) then 
              dummy = 1d-90
           endif
           if(index.gt.0) gasdens(index) = dummy
           call amr_nextcell(index)
        enddo
        close(1)
     else
        !
        ! Unformatted input of gas temperature
        !
        open(unit=1,file='gas_density.uinp',status='old',form='unformatted')
        read(1) iiformat,reclen
        if(iiformat.ne.1) then
           write(stdo,*) 'ERROR: Format number of gas_density.uinp is invalid/unknown.'
           stop
        endif
        reclend=reclen/8
        read(1) nn
        if(nn.ne.nrcellsinp) then
           write(stdo,*) 'ERROR: dust_density.uinp does not have same number'
           write(stdo,*) '       of cells as the grid.'
           write(stdo,*) nn,nrcellsinp
           stop
        endif
        if(allocated(data)) deallocate(data)
        allocate(data(1:reclend),STAT=ierr)
        if(ierr.ne.0) then
           write(stdo,*) 'ERROR: Could not allocate data array'
           stop
        endif
        call amr_resetcount()
        call amr_nextcell(index)
        do irec=1,(nrcellsinp-1)/reclend+1
           read(1) data
           i = 1
           do while((index.ge.0).and.(i.le.reclend))
              if(index.gt.0) gasdens(index) = data(i) 
              i     = i + 1
              call amr_nextcell(index)
           enddo
        enddo
        close(1)
     endif
  endif
end subroutine read_gas_density


!-------------------------------------------------------------------------
!                       READ GAS TEMPERATURE 
!-------------------------------------------------------------------------
subroutine read_gas_temperature(action)
  implicit none
  integer :: action
  logical :: fex1,fex2
  double precision, allocatable :: data(:)
  double precision :: dummy
  integer(kind=8) :: iiformat,reclen,reclend,nn,kk
  integer :: icell,i,ierr,iformat,irec,n,index
  !
  ! Action=0 means do nothing, action=1 means read if not yet read,
  ! action=2 means re-read.
  !
  if(action.eq.0) then
     return
  elseif(action.eq.1) then
     if(allocated(gastemp)) then
        !
        ! The gas temperature is already determined. It might be
        ! that also the gastmax has been calculated from
        ! this, but we don't know for sure. So let's calculate
        ! it here to make sure gastmax is up-to-date. It
        ! can be important if the gas temperature is created 
        ! using the userdef module.
        !
        ! Thanks, Rainer Rolffs, for pointing this out!
        ! 02.01.2011
        !
        gastmax = 0.d0
        do icell=1,nrcells
           index = cellindex(icell)
           if(gastemp(index).gt.gastmax) then
              gastmax = gastemp(index)
           endif
        enddo
        return
     endif
  endif
  ! 
  ! If the user requests, then use one of the dust temperatures as gas
  ! temperature
  !
  if(tgas_eq_tdust.gt.0) then
     !
     ! Use dust temperature
     !
     ! First make sure the dust data are read
     ! if they are not already read yet
     !
     call read_dustdata(1)
     call read_dust_density(1)
     call read_dust_temperature(1)
     !
     ! Check something
     !
     if(tgas_eq_tdust.gt.dust_nr_species) then
        write(stdo,*) 'ERROR: tgas_eq_tdust is larger than dust_nr_species'
        stop
     endif
     !
     ! Allocate array
     !
     if(allocated(gastemp)) deallocate(gastemp)
     allocate(gastemp(1:nrcells),STAT=ierr)
     if(ierr.ne.0) then
        write(stdo,*) 'ERROR: Could not allocate gastemp array'
        stop
     endif
     write(stdo,*) 'Using dust temperature as gas temperature...'
     call flush(stdo)
     !
     ! Now go and map the temperature 
     !
     do icell=1,nrcells
        index = cellindex(icell)
        if((index.le.0).or.(index.gt.nrcellsmax)) then
           write(stdo,*) 'ERROR: Internal error while reading gas temperature'
           stop
        endif
        gastemp(index) = dusttemp(tgas_eq_tdust,index)
     enddo
  else
     !
     ! Open temperature file
     !
     inquire(file='gas_temperature.inp',exist=fex1)
     inquire(file='gas_temperature.uinp',exist=fex2)
     if(fex1.and.fex2) then
        write(stdo,*) 'ERROR: Found both gas_temperature.inp and gas_temperature.uinp'
        stop
     endif
     if(fex1.or.fex2) then
        !
        ! Allocate array
        !
        if(allocated(gastemp)) deallocate(gastemp)
        allocate(gastemp(1:nrcells),STAT=ierr)
        if(ierr.ne.0) then
           write(stdo,*) 'ERROR: Could not allocate gastemp array'
           stop
        endif
        write(stdo,*) 'Reading gas temperature...'
        call flush(stdo)
        !
        ! Now read this temperature
        !
        if(fex1) then
           !
           ! Formatted gas temperature file
           !
           open(unit=1,file='gas_temperature.inp',status='old')
           read(1,*) iformat
           if(iformat.ne.1) then
              write(stdo,*) 'ERROR: Format number of gas_temperature.inp is invalid/unknown.'
              stop
           endif
           !read(1,*) gridfilenamedum
           read(1,*) n
           if(n.ne.nrcellsinp) then
              write(stdo,*) 'ERROR: gas_temperature.inp does not have same number'
              write(stdo,*) '       of cells as the grid.'
              write(stdo,*) n,nrcellsinp
              stop
           endif
           call amr_resetcount()
           call amr_nextcell(index)
           do while(index.ge.0)
              read(1,*) dummy 
              if(dummy.lt.0) then 
                 write(stdo,*) 'ERROR: Negative temperature detected in input file.'
                 stop
              endif
              if(dummy.lt.1d-90) then 
                 dummy = 1d-90
              endif
              if(index.gt.0) gastemp(index) = dummy
              call amr_nextcell(index)
           enddo
           close(1)
        else
           !
           ! Unformatted input of gas temperature
           !
           open(unit=1,file='gas_temperature.uinp',status='old',form='unformatted')
           read(1) iiformat,reclen
           if(iiformat.ne.1) then
              write(stdo,*) 'ERROR: Format number of gas_temperature.uinp is invalid/unknown.'
              stop
           endif
           reclend=reclen/8
           read(1) nn
           if(nn.ne.nrcellsinp) then
              write(stdo,*) 'ERROR: gas_temperature.uinp does not have same number'
              write(stdo,*) '       of cells as the grid.'
              write(stdo,*) nn,nrcellsinp
              stop
           endif
           if(allocated(data)) deallocate(data)
           allocate(data(1:reclend),STAT=ierr)
           if(ierr.ne.0) then
              write(stdo,*) 'ERROR: Could not allocate data array'
              stop
           endif
           call amr_resetcount()
           call amr_nextcell(index)
           do irec=1,(nrcellsinp-1)/reclend+1
              read(1) data
              i = 1
              do while((index.ge.0).and.(i.le.reclend))
                 if(index.gt.0) gastemp(index) = data(i) 
                 i     = i + 1
                 call amr_nextcell(index)
              enddo
           enddo
           close(1)
        endif
     else
        write(stdo,*) 'ERROR: Could not find gas_temperature.inp or gas_temperature.uinp.'
        stop
     endif
  endif
  !
  ! Compute the maximum temperature
  !
  gastmax = 0.d0
  do icell=1,nrcells
     index = cellindex(icell)
     if(gastemp(index).gt.gastmax) then
        gastmax = gastemp(index)
     endif
  enddo
  !
end subroutine read_gas_temperature


!-------------------------------------------------------------------------
!                   READ ELECTRON NUMBER DENSITY
!-------------------------------------------------------------------------
subroutine read_electron_numberdensity(action)
  implicit none
  integer :: action
  logical :: fex1,fex2
  double precision, allocatable :: data(:)
  double precision :: dummy
  integer(kind=8) :: iiformat,reclen,reclend,nn,kk
  integer :: i,index,iformat,irec,n,k,ierr
  !
  ! Action=0 means do nothing, action=1 means read if not yet read,
  ! action=2 means re-read.
  !
  if(action.eq.0) then
     return
  elseif(action.eq.1) then
     if(allocated(electron_numdens)) return
  endif
  !
  ! Open temperature file
  !
  inquire(file='electron_numdens.inp',exist=fex1)
  inquire(file='electron_numdens.uinp',exist=fex2)
  if(fex1.and.fex2) then
     write(stdo,*) 'ERROR: Found both electron_numdens.inp and electron_numdens.uinp'
     stop
  endif
  if(fex1.or.fex2) then
     !
     ! Allocate array
     !
     if(allocated(electron_numdens)) deallocate(electron_numdens)
     allocate(electron_numdens(1:nrcells),STAT=ierr)
     if(ierr.ne.0) then
        write(stdo,*) 'ERROR: Could not allocate electron_numdens array'
        stop
     endif
     write(stdo,*) 'Reading electron number density...'
     call flush(stdo)
     !
     ! Now read this temperature
     !
     if(fex1) then
        !
        ! Formatted electron number density temperature file
        !
        open(unit=1,file='electron_numdens.inp',status='old')
        read(1,*) iformat
        if(iformat.ne.1) then
           write(stdo,*) 'ERROR: Format number of electron_numdens.inp is invalid/unknown.'
           stop
        endif
        !read(1,*) gridfilenamedum
        read(1,*) n
        if(n.ne.nrcellsinp) then
           write(stdo,*) 'ERROR: electron_numdens.inp does not have same number'
           write(stdo,*) '       of cells as the grid.'
           write(stdo,*) n,nrcellsinp
           stop
        endif
        call amr_resetcount()
        call amr_nextcell(index)
        do while(index.ge.0)
           read(1,*) dummy 
           if(dummy.lt.0) then 
              write(stdo,*) 'ERROR: Negative electrons numder density detected in input file.'
              stop
           endif
           if(dummy.lt.1d-90) then 
              dummy = 1d-90
           endif
           if(index.gt.0) electron_numdens(index) = dummy
           call amr_nextcell(index)
        enddo
        close(1)
     else
        !
        ! Unformatted input of electron numbder density temperature
        !
        open(unit=1,file='electron_numdens.uinp',status='old',form='unformatted')
        read(1) iiformat,reclen
        if(iiformat.ne.1) then
           write(stdo,*) 'ERROR: Format number of electron_numdens.uinp is invalid/unknown.'
           stop
        endif
        reclend=reclen/8
        read(1) nn
        if(nn.ne.nrcellsinp) then
           write(stdo,*) 'ERROR: electron_numdens.uinp does not have same number'
           write(stdo,*) '       of cells as the grid.'
           write(stdo,*) nn,nrcellsinp
           stop
        endif
        if(allocated(data)) deallocate(data)
        allocate(data(1:reclend),STAT=ierr)
        if(ierr.ne.0) then
           write(stdo,*) 'ERROR: Could not allocate data array'
           stop
        endif
        call amr_resetcount()
        call amr_nextcell(index)
        do irec=1,(nrcellsinp-1)/reclend+1
           read(1) data
           i = 1
           do while((index.ge.0).and.(i.le.reclend))
              if(index.gt.0) electron_numdens(index) = data(i) 
              i     = i + 1
              call amr_nextcell(index)
           enddo
        enddo
        close(1)
     endif
  endif
end subroutine read_electron_numberdensity


!-------------------------------------------------------------------------
!                   READ ION NUMBER DENSITY
!-------------------------------------------------------------------------
subroutine read_ion_numberdensity(action)
  implicit none
  integer :: action
  logical :: fex1,fex2
  double precision, allocatable :: data(:)
  double precision :: dummy
  integer(kind=8) :: iiformat,reclen,reclend,nn,kk
  integer :: i,index,iformat,irec,n,k,ierr
  !
  ! Action=0 means do nothing, action=1 means read if not yet read,
  ! action=2 means re-read.
  !
  if(action.eq.0) then
     return
  elseif(action.eq.1) then
     if(allocated(ion_numdens)) return
  endif
  !
  ! Open temperature file
  !
  inquire(file='ion_numdens.inp',exist=fex1)
  inquire(file='ion_numdens.uinp',exist=fex2)
  if(fex1.and.fex2) then
     write(stdo,*) 'ERROR: Found both ion_numdens.inp and ion_numdens.uinp'
     stop
  endif
  if(fex1.or.fex2) then
     !
     ! Allocate array
     !
     if(allocated(ion_numdens)) deallocate(ion_numdens)
     allocate(ion_numdens(1:nrcells),STAT=ierr)
     if(ierr.ne.0) then
        write(stdo,*) 'ERROR: Could not allocate ion_numdens array'
        stop
     endif
     write(stdo,*) 'Reading ion number density...'
     call flush(stdo)
     !
     ! Now read this temperature
     !
     if(fex1) then
        !
        ! Formatted ion number density temperature file
        !
        open(unit=1,file='ion_numdens.inp',status='old')
        read(1,*) iformat
        if(iformat.ne.1) then
           write(stdo,*) 'ERROR: Format number of ion_numdens.inp is invalid/unknown.'
           stop
        endif
        !read(1,*) gridfilenamedum
        read(1,*) n
        if(n.ne.nrcellsinp) then
           write(stdo,*) 'ERROR: ion_numdens.inp does not have same number'
           write(stdo,*) '       of cells as the grid.'
           write(stdo,*) n,nrcellsinp
           stop
        endif
        call amr_resetcount()
        call amr_nextcell(index)
        do while(index.ge.0)
           read(1,*) dummy 
           if(dummy.lt.0) then 
              write(stdo,*) 'ERROR: Negative ions numder density detected in input file.'
              stop
           endif
           if(dummy.lt.1d-90) then 
              dummy = 1d-90
           endif
           if(index.gt.0) ion_numdens(index) = dummy
           call amr_nextcell(index)
        enddo
        close(1)
     else
        !
        ! Unformatted input of ion numbder density temperature
        !
        open(unit=1,file='ion_numdens.uinp',status='old',form='unformatted')
        read(1) iiformat,reclen
        if(iiformat.ne.1) then
           write(stdo,*) 'ERROR: Format number of ion_numdens.uinp is invalid/unknown.'
           stop
        endif
        reclend=reclen/8
        read(1) nn
        if(nn.ne.nrcellsinp) then
           write(stdo,*) 'ERROR: ion_numdens.uinp does not have same number'
           write(stdo,*) '       of cells as the grid.'
           write(stdo,*) nn,nrcellsinp
           stop
        endif
        if(allocated(data)) deallocate(data)
        allocate(data(1:reclend),STAT=ierr)
        if(ierr.ne.0) then
           write(stdo,*) 'ERROR: Could not allocate data array'
           stop
        endif
        call amr_resetcount()
        call amr_nextcell(index)
        do irec=1,(nrcellsinp-1)/reclend+1
           read(1) data
           i = 1
           do while((index.ge.0).and.(i.le.reclend))
              if(index.gt.0) ion_numdens(index) = data(i) 
              i     = i + 1
              call amr_nextcell(index)
           enddo
        enddo
        close(1)
     endif
  endif
end subroutine read_ion_numberdensity


!-------------------------------------------------------------------------
!                  WRITE RESULTS OF BJORKMAN & WOOD
!-------------------------------------------------------------------------
subroutine write_results_mc_bjorkmanwood(ifile)
  implicit none
  integer :: ifile,ierror,ispec,index,inu,reclend,irec,i,ierr
  integer(kind=8) :: nn,kk
  logical :: fex,warnflag
  double precision, allocatable :: data(:)
  warnflag=.false.
  !
  ! For now we always use a single grid file
  !
  !gridfilename    = "gridfile.inp"
  !gridfilenamelen = 12
  !
  ! For now only formatted IO
  !
  !if(rto_unformatted) then
  !   write(stdo,*) 'ERROR: For now only the formatted I/O possible'
  !   stop 
  !endif
  !
  ! Write the grid file
  ! NOTE: For now I will assume that the grid has not been refined 
  !       during the simulation, and hence the grid file does not have
  !       to be rewritten. Also note for the future: if the grid file
  !       will be rewritten, then it must be to another file, and the
  !       dust_temperature.dat file must have an entry telling which
  !       grid file to use, so as not to mix up with the grid file for
  !       the input files such as dust_density.inp.
  !
  !!!! call write_grid_file()   
  !
  ! Now write the dust temperature
  !
  if(igrid_type.lt.100) then
     !
     ! Regular (AMR) grid
     ! 
     ! Just make sure that the cell list is complete
     !
     call amr_compute_list_all()
     !
     ! Do a stupidity check
     !
     if(nrcells.ne.amr_nrleafs) stop 3209
     !
     ! Open file and write the temperature to it
     !
     if(.not.rto_unformatted) then 
        !
        ! Write the temperatures in ascii form
        !
        open(unit=1,file='dust_temperature.dat')
        write(1,*) 1                                   ! Format number
        write(1,*) nrcellsinp
        write(1,*) dust_nr_species
        do ispec=1,dust_nr_species
           call amr_resetcount()
           call amr_nextcell(index)
           do while(index.ge.0)
              if(index.gt.0) then
                 write(1,*) dusttemp(ispec,index)
              else
                 write(1,*) '0'
              endif
              call amr_nextcell(index)
           enddo
        enddo
        close(1)
     else
        !
        ! Write the temperatures in f77-style unformatted form,
        ! using a record length given by rto_reclen
        !
        reclend = rto_reclen/8
        if(allocated(data)) deallocate(data)
        allocate(data(1:reclend),STAT=ierr)
        if(ierr.ne.0) then
           write(stdo,*) 'ERROR: Could not allocate data array'
           stop
        endif
        open(unit=1,file='dust_temperature.udat',form='unformatted')
        nn = 1
        kk = rto_reclen
        write(1) nn,kk               ! Format number and record length
        nn = nrcellsinp
        kk = dust_nr_species
        write(1) nn,kk
        do ispec=1,dust_nr_species
           call amr_resetcount()
           call amr_nextcell(index)
           do irec=1,(nrcellsinp-1)/reclend+1
              i = 1
              data(:) = 0.d0
              do while((index.ge.0).and.(i.le.reclend))
                 if(index.gt.0) then
                    data(i) = dusttemp(ispec,index)
                 else
                    data(i) = 0.d0
                 endif
                 i       = i + 1
                 call amr_nextcell(index)
              enddo
              write(1) data
           enddo
        enddo
        deallocate(data)
     endif
     !
     ! If scattering is isotropic, then open file and write that data
     !
!     if(scattering_mode.eq.1) then
!        if(.not.rto_unformatted) then 
!           !
!           ! Write the scattering source in ascii form
!           !
!           open(unit=1,file='scatsrc_iso.dat')
!           write(1,*) 1                                   ! Format number
!           write(1,*) amr_nrleafs
!           write(1,*) freq_nr
!           do inu=1,freq_nr
!              do icell=1,nrcells
!                 index = amr_theleaf_index(icell)
!                 write(1,*) isoscatsrc(inu,index)
!                 if((isoscatsrc(inu,index).lt.0.d0).and.(.not.warnflag)) then
!                    write(stdo,*) 'WARNING while writing scatsrc_iso.dat: Negative source found.'
!                    warnflag=.true.
!                 endif
!              enddo
!           enddo
!           close(1)
!        else
!           !
!           ! Write the scattering source in f77-style unformatted form,
!           ! using a record length given by rto_reclen
!           !
!           reclend = rto_reclen/8
!           if(allocated(data)) deallocate(data)
!           allocate(data(1:reclend),STAT=ierr)
!           if(ierr.ne.0) then
!              write(stdo,*) 'ERROR: Could not allocate data array'
!              stop
!           endif
!           open(unit=1,file='scatsrc_iso.udat',form='unformatted')
!           nn = 1
!           kk = rto_reclen
!           write(1) nn,kk               ! Format number and record length
!           nn = nrcells
!           kk = freq_nr
!           write(1) nn,kk
!           do inu=1,freq_nr
!              icell = 1
!              do irec=1,(nrcells-1)/reclend+1
!                 i = 1
!                 data(:) = 0.d0
!                 do while((icell.le.nrcells).and.(i.le.reclend))
!                    index = cellindex(icell)
!                    data(i) = isoscatsrc(inu,index)
!                    i       = i + 1
!                    icell   = icell + 1
!                 enddo
!                 write(1) data
!              enddo
!           enddo
!           deallocate(data)
!        endif
!     endif
  else
     !
     ! Other grids not yet implemented
     !
     write(stdo,*) 'ERROR: Only regular and AMR grids implemented'
     stop
  endif
  !
  ! If the grid is internally made, then we must make sure that
  ! the grid has been written to file, otherwise the output file
  ! created here makes no sense.
  !
  if((.not.grid_was_read_from_file).and.(.not.grid_was_written_to_file)) then
     call write_grid_file()
     grid_was_written_to_file = .true.     ! Avoid multiple writings
  endif
  !
end subroutine write_results_mc_bjorkmanwood



!-------------------------------------------------------------------------
!                    WRITE GRID FILE (UNIVERSAL)
!-------------------------------------------------------------------------
subroutine write_grid_file()
  implicit none
  character*160 :: filename
  logical :: formatted
  formatted = .not.rto_unformatted
  if(igrid_type.lt.100) then
     if(formatted) then
        filename = 'amr_grid.inp'
     else
        filename = 'amr_grid.uinp'
     endif
     write(stdo,*) 'Writing grid file to ',trim(filename)
     call amr_write_grid(filename,formatted,rtio_gridinfo)
  else
     write(stdo,*) 'ERROR: In write_grid_file(): unstructured grids not yet supported'
     stop
  endif
end subroutine write_grid_file


!-------------------------------------------------------------------------
!                    READ GRID FILE (UNIVERSAL)
!-------------------------------------------------------------------------
subroutine read_grid_file()
  implicit none
  character*160 :: filename
  logical :: formatted,fex_amrgrid,fex_delaunaygrid,fex_voronoigrid
  logical :: fex1,fex2
  integer :: amr,delaunay,voronoi,icell,idir
  double precision :: dd
  !
  ! Check out which grid file is present
  !
  inquire(file='amr_grid.inp',exist=fex1)
  inquire(file='amr_grid.uinp',exist=fex2)
  fex_amrgrid=fex1.or.fex2
  inquire(file='triang_vertex_grid.inp',exist=fex1)
  inquire(file='triang_vertex_grid.uinp',exist=fex2)
  fex_delaunaygrid=fex1.or.fex2
  inquire(file='voronoi_grid.inp',exist=fex1)
  inquire(file='voronoi_grid.inp',exist=fex2)
  fex_voronoigrid=fex1.or.fex2
  if(fex_amrgrid) then
     amr=2 
  else 
     amr=1
  endif
  if(fex_delaunaygrid) then
     delaunay=2 
  else 
     delaunay=1
  endif
  if(fex_voronoigrid) then
     voronoi=2 
  else 
     voronoi=1
  endif
  if(amr*delaunay*voronoi.gt.2) then
     write(stdo,*) 'ERROR: More than one (amr/delaunay/voronoi)_grid.inp file found.'
     write(stdo,*) '       You must choose which kind of grid you want to use.'
     stop
  elseif(amr*delaunay*voronoi.eq.1) then
     write(stdo,*) 'ERROR: No file (amr/delaunay/voronoi)_grid.inp found.'
     write(stdo,*) '       I need this file to set up a grid.'
     stop
  endif
  !
  ! For each grid type call the appropriate routine. This automatically
  ! initializes the grid and makes the linear list of cells that is 
  ! necessary for storing data. For the AMR grid this is the list that is
  ! computed by amr_compute_list_all(). The order in which the leafs (=cells)
  ! are listed in this list is also the order in which their cells are 
  ! read/written.
  !
  ! NOTE: In the future, for extremely large simulations that do not fit 
  !       onto a single node we may be forced to do this differently. We
  !       would then (in the AMR grid) have partial trees. Still, the 
  !       order of the leafs would still the same as the order of the 
  !       list. 
  !
  if(fex_amrgrid) then
     inquire(file='amr_grid.inp',exist=fex1)
     inquire(file='amr_grid.uinp',exist=fex2)
     if(fex1.and.fex2) then
        write(stdo,*) 'ERROR: Found both amr_grid.inp and amr_grid.uinp'
        stop
     endif
     if((.not.fex1).and.(.not.fex2)) then
        write(stdo,*) 'ERROR: Could not find the amr_grid.inp or amr_grid.uinp file...'
        stop
     endif
     if(fex1) then
        filename  = 'amr_grid.inp'
        formatted = .true.
     else
        filename  = 'amr_grid.uinp'
        formatted = .false.
     endif
     call amr_read_grid(filename,formatted,checkspher=.true.)
     igrid_type = amr_style
     if(igrid_type.ge.100) stop 4091
     if(igrid_type.lt.0) stop 4092
  else
     write(stdo,*) 'ERROR: In read_grid_file(): unstructured grids not yet supported'
     stop
  endif
  !
  ! Remember that we read this grid from a file, i.e. we did NOT 
  ! generate the grid internally in the code. This means that when 
  ! writing results of e.g. the Monte Carlo simulation, we do not
  ! need to write the grid again.
  !
  grid_was_read_from_file = .true.
  !
end subroutine read_grid_file


!-------------------------------------------------------------------------
!                        POST PROCESS GRID
!-------------------------------------------------------------------------
subroutine postprocess_grid_for_use(renew)
  implicit none
  double precision :: dd
  integer :: i,ierr,idum,idir,icell,ilev
  integer :: ix,iy,iz,istar
  logical, optional :: renew
  logical :: dorenew
  character*80 :: str1,str2
  double precision :: dummy
  !
  ! Check optional argument
  ! 
  if(present(renew)) then 
     dorenew = renew 
  else
     dorenew = .false.
  endif
  !
  ! If dorenew then unallocate arrays
  !
  if(dorenew) then
     if(allocated(cellindex)) deallocate(cellindex)
     if(allocated(cellvolume)) deallocate(cellvolume)
     nrcells    = 0
     nrcellsmax = 0
  endif
  !
  ! Switch between grid forms
  !
  if((igrid_type.ge.0).and.(igrid_type.lt.100)) then
     !
     ! For backward compatibility, redefine igrid_type to be 
     ! the same as the amr_style
     !
     igrid_type = amr_style
     !
     ! Check that the coordinates are always increasing
     !
     do ix=2,amr_grid_nx+1
        if(amr_grid_xi(ix,1).le.amr_grid_xi(ix-1,1)) then
           write(stdo,*) 'ERROR: coordinates must always increase!'
           stop
        endif
     enddo
     do iy=2,amr_grid_ny+1
        if(amr_grid_xi(iy,2).le.amr_grid_xi(iy-1,2)) then
           write(stdo,*) 'ERROR: coordinates must always increase!'
           stop
        endif
     enddo
     do iz=2,amr_grid_nz+1
        if(amr_grid_xi(iz,3).le.amr_grid_xi(iz-1,3)) then
           write(stdo,*) 'ERROR: coordinates must always increase!'
           stop
        endif
     enddo
     !
     ! Import the coordinate system number
     !
     igrid_coord = amr_coordsystem
     !
     ! Check the grid
     !
     if(igrid_coord.lt.100) then
        !
        ! Cartesian grid
        !
        ! Just for safety, check that the user does not use a coordinate
        ! system (in case of Cartesian coordinates) in which the (0,0,0)
        ! point is very very far outside of the domain. That could give
        ! trouble in the imaging routines of the camera_module, because
        ! some things are not calculated with translational invariance.
        !
        if((abs(amr_grid_xi(amr_grid_nx+1,1)-amr_grid_xi(1,1)).lt.                &
             1d-4*max(abs(amr_grid_xi(amr_grid_nx+1,1)),abs(amr_grid_xi(1,1)))).or. &
             (abs(amr_grid_xi(amr_grid_ny+1,2)-amr_grid_xi(1,2)).lt.                &
             1d-4*max(abs(amr_grid_xi(amr_grid_ny+1,2)),abs(amr_grid_xi(1,2)))).or. &
             (abs(amr_grid_xi(amr_grid_nz+1,3)-amr_grid_xi(1,3)).lt.                &
             1d-4*max(abs(amr_grid_xi(amr_grid_nz+1,3)),abs(amr_grid_xi(1,3))))) then
           write(stdo,*) 'ERROR: Please put computational domain closer'
           write(stdo,*) '       to the (0,0,0) point of the coordinates'
           write(stdo,*) '       to avoid numerical accuracy problems.'
           stop
        endif
        !
        ! If using Cartesian AMR gridding, then we must insist that the
        ! base cells are cubic and all of the same size. This is because
        ! the camera_module.f90 imaging routines can only find the proper
        ! recursive refinement of the image pixels if this is so. This
        ! recursive refinement is important for ensuring that all the flux
        ! is picked up properly. 
        !
        dd  = amr_grid_xi(2,1) - amr_grid_xi(1,1)
        if(dd.le.0.d0) then
           write(stdo,*) 'ERROR while reading grid: base grid is bad.'
           stop
        endif
        do i=1,amr_grid_nx
           if(abs((amr_grid_xi(i+1,1)-amr_grid_xi(i,1))/dd-1.d0).gt.1d-4) then
              write(stdo,*) 'ERROR while reading grid: base grid is not regularly cubically spaced.'
              stop
           endif
        enddo
        do i=1,amr_grid_ny
           if(abs((amr_grid_xi(i+1,2)-amr_grid_xi(i,2))/dd-1.d0).gt.1d-4) then
              write(stdo,*) 'ERROR while reading grid: base grid is not regularly cubically spaced.'
              stop
           endif
        enddo
        do i=1,amr_grid_nz
           if(abs((amr_grid_xi(i+1,3)-amr_grid_xi(i,3))/dd-1.d0).gt.1d-4) then
              write(stdo,*) 'ERROR while reading grid: base grid is not regularly cubically spaced.'
              stop
           endif
        enddo
        !
        ! Compute grid center and the radius of the containing sphere.
        ! These variables are useful for various Monte Carlo things, 
        ! including the external luminosity, and improved photon statistics
        ! for externally located stars.
        !
        grid_contsph_x = 0.5d0 * ( amr_grid_xi(amr_grid_nx+1,1) + amr_grid_xi(1,1) )
        grid_contsph_y = 0.5d0 * ( amr_grid_xi(amr_grid_ny+1,2) + amr_grid_xi(1,2) )
        grid_contsph_z = 0.5d0 * ( amr_grid_xi(amr_grid_nz+1,3) + amr_grid_xi(1,3) )
        grid_contsph_r = 0.5001d0 * sqrt( ( amr_grid_xi(amr_grid_nx+1,1) - amr_grid_xi(1,1) )**2 +  &
                                          ( amr_grid_xi(amr_grid_ny+1,2) - amr_grid_xi(1,2) )**2 +  &
                                          ( amr_grid_xi(amr_grid_nz+1,3) - amr_grid_xi(1,3) )**2 )
        !   
     elseif(igrid_coord.lt.200) then
        !
        ! Spherical coordinates
        !
        ! Check that the grid is always monotonically increasing
        ! 
        do i=1,amr_grid_nx
           if(amr_grid_xi(i+1,1).le.amr_grid_xi(i,1)) then
              write(stdo,*) 'ERROR while reading grid: base R grid is not monotonically increasing.'
              stop
           endif
        enddo
        do i=1,amr_grid_ny
           if(amr_grid_xi(i+1,2).le.amr_grid_xi(i,2)) then
              write(stdo,*) 'ERROR while reading grid: base Theta grid is not monotonically increasing.'
              stop
           endif
        enddo
        do i=1,amr_grid_nz
           if(amr_grid_xi(i+1,3).le.amr_grid_xi(i,3)) then
              write(stdo,*) 'ERROR while reading grid: base Phi grid is not monotonically increasing.'
              stop
           endif
        enddo
        !
        ! Compute grid center and the radius of the containing sphere.
        ! These variables are useful for various Monte Carlo things, 
        ! including the external luminosity, and improved photon statistics
        ! for externally located stars.
        !
        grid_contsph_x = 0.d0
        grid_contsph_y = 0.d0
        grid_contsph_z = 0.d0
        grid_contsph_r = 1.001 * amr_grid_xi(amr_grid_nx+1,1)
        !
     else
        stop 609
     endif
     !
     ! Because the radiative transfer code should be as independent as 
     ! possible of the details of the AMR grid (so that other types of
     ! grids can easily be included too) we make our own independent
     ! indexing array here and have some other copies of variables.
     !
     ! NOTE (bugfix 28.07.09): Here we must allocate the arrays with
     !      nrcellsmax size instead of nrcells. The reason is that 
     !      the user MAY want to refine the grid lateron.
     !
     nrcells     = amr_nrleafs
     !
     ! Now set the nrcellsinp value, which is the counter for how
     ! many cells will be read. This is equal to nrcells for regular
     ! and AMR grids of the oct-tree type. But for the layer-style
     ! AMR there is redundancy, to keep the data format easy.
     !
     if(amr_style.lt.10) then
        nrcellsinp = nrcells
     elseif(amr_style.lt.20) then
        nrcellsinp = amr_nrbranches
     else
        write(stdo,*) 'ERROR: AMR grid style ',amr_style,' does not exist.'
        stop
     endif
     !
     ! Allocate the arrays only if they have not yet been allocated
     !
     if(.not.allocated(cellindex)) then 
        !
        ! Not yet allocated, so allocate
        !
        !nrcellsmax  = amr_nrleafs       ! bugfix 28.07.09
        nrcellsmax  = amr_nrleafs_max
        ! Bug? 16.11.09: Why do I allocate with nrcells and not with nrcellsmax?
        allocate(cellindex(1:nrcells),STAT=ierr)
        if(ierr.ne.0) then
           write(stdo,*) 'ERROR: Could not allocate cellindex array'
           stop
        endif
        do icell=1,nrcells
           cellindex(icell) = amr_theleaf_index(icell)
        enddo
        ! Bug? 16.11.09: Why do I allocate with nrcells and not with nrcellsmax?
        allocate(cellvolume(1:nrcells),STAT=ierr) 
        if(ierr.ne.0) then
           write(stdo,*) 'ERROR: Could not allocate cellvolume array'
           stop
        endif
     else
        !
        ! Already allocated, so just do some checks
        !
        if(nrcellsmax.ne.amr_nrleafs_max) then
           write(stdo,*) 'ERROR: cellindex already allocated, but with '
           write(stdo,*) '       different array size...'
           stop
        endif
     endif
     !
     ! Compute the cell volumes
     !
     call compute_cellvolumes_amr()
     !
     ! Calculate the ray_nsmax
     ! The factor of 2 is necessary for spherical or cylindrical coordinates.
     !
     ray_nsmax = 0
     do idir=1,3
        idum = 2 * amr_grid_nx * 2**amr_levelmax + 1
        ray_nsmax = max(ray_nsmax,idum)
     enddo
     !
     ! Do a message
     !
     ! NOTE: Since we changed the levels to 0:amr_levelmax (from originally
     !       1:amr_levelmax), we must also start from 0 here, contrary to
     !       earlier versions.
     !
     write(stdo,*) 'Grid information (current status):'
     write(str1,*) amr_nrbranches
     write(str2,*) amr_nrleafs
     write(stdo,*) '  We have ',trim(adjustl(str1)),' branches, of which ', &
          trim(adjustl(str2)),' are actual grid cells.'
     write(str1,321) 100*amr_nrbranches/(1.d0*amr_nrbranches_max)
     write(str2,321) 100*amr_nrleafs/(1.d0*amr_nrleafs_max)
321 format(F7.3)
     write(stdo,*) '  ---> ',trim(adjustl(str1)),'% mem use for branches, and ', &
          trim(adjustl(str2)),'% mem use for actual cells.'
     ilev = 0
     do icell=1,nrcells
        if(ilev.lt.amr_theleafs(icell)%link%level) then
           ilev = amr_theleafs(icell)%link%level
        endif
     enddo
     if(ilev.eq.0) then
        write(stdo,*) '  No grid refinement is active (at present; userdef may change this).'
     else
        write(str1,*) ilev
        write(stdo,*) '  Top refinement level of the grid = ',trim(adjustl(str1))
     endif
     !
  elseif((igrid_type.ge.100).and.(igrid_type.lt.200)) then
     write(stdo,*) 'ERROR: In postprocess_grid_for_use(): unstructured grids not yet supported'
     stop
  else
     write(stdo,*) 'ERROR: In postprocess_grid_for_use(): igrid_type ',igrid_type,' not known.'
     stop
  endif
  !
  ! Compute the farthest distance of a star from the edge of the 
  ! containing sphere
  !
  star_maxexterndist = 0.d0
  if((nstars.gt.0).and.(grid_contsph_r.gt.0.d0)) then
     do istar=1,nstars
        dummy = sqrt((star_pos(1,istar)-grid_contsph_x)**2+&
                     (star_pos(2,istar)-grid_contsph_y)**2+&
                     (star_pos(3,istar)-grid_contsph_z)**2)
        if(dummy.gt.grid_contsph_r) then
           if(dummy-grid_contsph_r+star_r(istar).gt.star_maxexterndist) then
              star_maxexterndist = dummy-grid_contsph_r+star_r(istar)
           endif
        endif
     enddo
  endif
  !
end subroutine postprocess_grid_for_use



!-------------------------------------------------------------------------
!           COMPUTE ALL CELL VOLUMES AND INDICES OF AMR GRID
!-------------------------------------------------------------------------
subroutine compute_cellvolumes_amr()
  implicit none
  !
  integer :: ileaf,leafindex,iddr
  type(amr_branch), pointer :: leaf
  doubleprecision :: axi(1:2,1:3)
  !
  ! Check if we can use the leaf list of the AMR module
  !
  if(.not.allocated(amr_theleafs)) then
     write(stdo,*) 'ERROR while computing cell volume: The leaf list is not yet made.'
     stop
  endif
  !
  ! Now loop over all leafs
  !
  if(igrid_coord.lt.100) then  
     !
     ! Cartesian coordinates
     !
     do ileaf=1,amr_nrleafs
        leafindex = amr_theleaf_index(ileaf)
        leaf      => amr_theleafs(ileaf)%link
        do iddr=1,3
           axi(1,iddr) = amr_finegrid_xi(leaf%ixyzf(iddr),iddr,leaf%level)
           axi(2,iddr) = amr_finegrid_xi(leaf%ixyzf(iddr)+1,iddr,leaf%level)
        enddo
        cellvolume(leafindex) = ( axi(2,1) - axi(1,1) ) *   &
                                ( axi(2,2) - axi(1,2) ) *   &
                                ( axi(2,3) - axi(1,3) )
     enddo
  elseif(igrid_coord.lt.200) then  
     !
     ! Spherical coordinates
     !
     do ileaf=1,amr_nrleafs
        leafindex = amr_theleaf_index(ileaf)
        leaf      => amr_theleafs(ileaf)%link
        do iddr=1,3
           axi(1,iddr) = amr_finegrid_xi(leaf%ixyzf(iddr),iddr,leaf%level)
           axi(2,iddr) = amr_finegrid_xi(leaf%ixyzf(iddr)+1,iddr,leaf%level)
        enddo
        cellvolume(leafindex) = (1.d0/3.d0)*                  &
             ( axi(2,1)**3 - axi(1,1)**3 ) *                  &
             ( cos(axi(1,2)) - cos(axi(2,2)) ) *              &
             ( axi(2,3) - axi(1,3) )
     enddo
  else
     stop 823
  endif
  !
end subroutine compute_cellvolumes_amr


!-------------------------------------------------------------------------
!                      COMPUTE DUST MASS
!-------------------------------------------------------------------------
subroutine compute_dust_mass()
  use stars_module
  use amr_module
  implicit none
  integer :: ileaf,ispec,leafindex
  !
  ! Check if we can use the leaf list of the AMR module
  !
  if(.not.allocated(amr_theleafs)) then
     write(stdo,*) 'ERROR while computing dust mass: The leaf list is not yet made.'
     stop
  endif
  !
  ! Reset the dust masses
  !
  do ispec=1,dust_nr_species
     dust_massdust(ispec) = 0.d0
  enddo
  !
  ! Now do a loop over the leafs and dust species
  !
  do ileaf=1,amr_nrleafs
     leafindex = amr_theleaf_index(ileaf)
     do ispec=1,dust_nr_species
        dust_massdust(ispec) = dust_massdust(ispec) +                 &
             cellvolume(leafindex)*dustdens(ispec,leafindex)
     enddo
  enddo
  !
  ! Now compute total dust mass
  ! 
  dust_massdusttot = 0.d0
  do ispec=1,dust_nr_species
     dust_massdusttot = dust_massdusttot + dust_massdust(ispec)
  enddo
  !
end subroutine compute_dust_mass


!-------------------------------------------------------------------------
!                        WRITE DUST DENSITY
!-------------------------------------------------------------------------
subroutine write_dust_density()
  implicit none
  double precision, allocatable :: data(:)
  integer(kind=8) :: iiformat,reclend,nn,kk
  integer :: iformat,index,ierr,irec,ispec,i
  character*160 :: filename
  logical :: formatted
  !
  formatted = .not.rto_unformatted
  if(formatted) then
     filename = 'dust_density.inp'
  else
     filename = 'dust_density.uinp'
  endif
  write(stdo,*) 'Writing dust densities to ',trim(filename)
  call flush(stdo)
  if(formatted) then
     !
     ! Formatted output
     !
     open(unit=1,file=filename)
     write(1,*) 1
     write(1,*) nrcellsinp
     write(1,*) dust_nr_species
     do ispec=1,dust_nr_species
        call amr_resetcount()
        call amr_nextcell(index)
        do while(index.ge.0)
           if(index.gt.0) then
              write(1,*) dustdens(ispec,index)
           else
              write(1,*) '0'
           endif
           call amr_nextcell(index)
        enddo
     enddo
  else
     !
     ! Unformatted output
     !
     reclend = rto_reclen/8
     open(unit=1,file=filename,form='unformatted')
     nn=1
     kk=rto_reclen
     write(1) nn,kk
     nn=nrcellsinp
     kk=dust_nr_species
     write(1) nn,kk
     if(allocated(data)) deallocate(data)
     allocate(data(1:reclend),STAT=ierr)
     if(ierr.ne.0) then
        write(stdo,*) 'ERROR: Could not allocate data array'
        stop
     endif
     do ispec=1,dust_nr_species
        call amr_resetcount()
        call amr_nextcell(index)
        do irec=1,(nrcellsinp-1)/reclend+1
           data(:) = 0.d0
           i = 1
           do while((index.ge.0).and.(i.le.reclend))
              if(index.gt.0) then
                 data(i) = dustdens(ispec,index)
              else
                 data(i) = 0.d0
              endif
              i     = i + 1
              call amr_nextcell(index)
           enddo
           write(1) data
        enddo
     enddo
  endif
  close(1)
  if(allocated(data)) deallocate(data)
  !
end subroutine write_dust_density


!--------------------------------------------------------------------------
!                            Main subbox routine 
!--------------------------------------------------------------------------
subroutine do_subbox_extract_and_write(nv,nc,subboxfilename,           &
     nx,ny,nz,x0,x1,y0,y1,z0,z1,phi1,theta,phi2,func)
  implicit none
  integer :: nc,nx,ny,nz,nv,ix,iy,iz,iv
  double precision :: x0,x1,y0,y1,z0,z1,phi1,theta,phi2
  double precision :: func(nv,nc)
  double precision :: funcslice(nx,ny,nz,nv)
  character*160 :: subboxfilename
  !
  ! Get the sub box
  !
  if(igrid_type.lt.100) then
     if(igrid_coord.lt.100) then
        call amr_subbox_3d(nv,nc,nx,ny,nz,x0,x1,y0,y1,z0,z1,           &
                           phi1,theta,phi2,func,funcslice)
     elseif(igrid_coord.lt.200) then
        call amr_subbox_3d(nv,nc,nx,ny,nz,x0,x1,y0,y1,z0,z1,           &
                  phi1,theta,phi2,func,funcslice,spherical=.true.)
     else
        stop 5092
     endif
  else
     write(stdo,*) 'ERROR: No other grid types than regular are allowed yet.'
     stop
  endif
  !
  ! Now write to a file
  !
  open(unit=1,file=trim(subboxfilename))
  write(1,*) 1                         ! Format number
  write(1,*) nx,ny,nz,nv               ! Dimensions
  write(1,*) x0,x1,y0,y1,z0,z1         ! Size and position of box
  write(1,*) phi1,theta,phi2           ! Orientation of the box
  write(1,*) 
  do iv=1,nv
     do iz=1,nz
        do iy=1,ny
           do ix=1,nx
              write(1,*) funcslice(ix,iy,iz,iv)
           enddo
        enddo
     enddo
  enddo
  close(1)
end subroutine do_subbox_extract_and_write




end module ioput_module
